/**
Solution to du/dt + du/dx = 0 using finite differences
and DEVS integrator with initial data a triangle
*/
#include <fstream>
#include <iostream>
#include <math.h>
#include <stdlib.h>
#include "adevs.h"
using namespace std;

class cell: public atomic, public cell_interface
{
	public:
		// Spatial discretization
		static const double k;
		// Integration quantum
		static double D;

		/// Initialize with the left value and center value.
		cell(double ul = 0.0, double u = 0.0):
		atomic(),
		q(),
		ul(ul)
		{
			q.setParams(D);
			double du = -(u-ul)/k;
			q.init(&u,&du);
		}
		cell(const cell& src):
		atomic(src),
		cell_interface(src),
		q(src.q),
		ul(src.ul)
		{
		}
		const cell& operator=(const cell& src)
		{
			atomic::operator=(src);
			cell_interface::operator=(src);
			q = src.q;
			ul = src.ul;
			return *this;
		}
		double get_state() const 
		{
			return q.state();
		}
		void init()
		{
			hold(q.ta());
		}
		void delta_int()
		{
			double du = -(q.out()-ul)/k;
			q.integ(q.ta(),&du);
			hold(q.ta());
		}
		void delta_ext(double e, const adevs_bag<PortValue>& x)
		{
			for (int i = 0; i < x.getSize(); i++)
			{
				if (x.get(i).port == cell_interface::in[0][1][1])
				{
					ul = x.get(i).value;
				}
			}
			q.integ(e);
			double du = -(q.qstate()-ul)/k;
			q.integ(0.0,&du);
			hold(q.ta());
		}
		void delta_conf(const adevs_bag<PortValue>& x)
		{
			delta_ext(q.ta(),x);
		}
		void output_func(adevs_bag<PortValue>& y)
		{
			output(out,q.out(),y);
		}
		void gc_output(adevs_bag<PortValue>& g)
		{
		}
		~cell()
		{
		}
	private:
		qoden<1> q;
		double ul;
};

// Quantum size for cell integrator
double cell::D = 0.1;
// Resolution in space
const double cell::k = 0.1;
// Number of cells in interval [0,1] given discretization k
static const int cell_count = (int)ceil(1.0/cell::k)+1;
// Time over which to compute the solution
static const double tend = 0.5;

int main(int argc, char** argv)
{
	cellSpace space(cellSpace::SIX_POINT,cell_count);
	for (int i = 1; i < cell_count; i++)
	{
		/// Triangular initial condition on interval [0.0,0.5] with
		/// peak height of 0.1
		double u = i*cell::k;
		double ul = (i-1)*cell::k;
		if (i*cell::k > 0.25) u = 0.5-i*cell::k;
		if ((i-1)*cell::k > 0.25) ul = 0.5-(i-1)*cell::k;
		if (i*cell::k >= 0.5) u = 0.0;
		if ((i-1)*cell::k >= 0.5) ul = 0.0;
		space.add(new cell(ul,u),i-1);
	}
	devssim sim(&space);
	sim.run(tend);
	for (int i = 0; i < cell_count-1; i++)
	{
		const cell* c = dynamic_cast<const cell*>(space.getModel(i));
		cout << i*cell::k << " " << c->get_state() << endl;
	}
	return 0;
}
